{% include "header.html" %}

<nav aria-label="breadcrumb">
  <ol class="breadcrumb bg-light">
    {% for crumb in crumbs %}
      {% if not loop.last %}
    <li class="breadcrumb-item"><a href="{{ crumb.url }}/">{{ crumb.part }}</a></li>
      {% else %}<li class="breadcrumb-item active" aria-current="page">{{ crumb.part }}</li>
      {% endif %}
    {% endfor %}

    <li class="ml-auto json-link"><a target="_blank" href="{{ url }}?f=schemajson">JSON</a></li>
  </ol>
</nav>

<h1>Collection: {{ response.title or response.id }}</h1>

<div class="row">
  <div class="col-sm">
    <h2>Queryables</h2>
    <ul>
    {% for k,v in response.properties.items() %}
      <li>{% if '$ref' in v %}
          <a href="{{ v['$ref'] }}">{{ k }}</a>
          {% else %}
          {{ k }}: {{ v['type'] }}
          {% endif %}
    {% endfor %}
    </ul>
  </div>
</div>

{% include "footer.html" %}
